// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_BINARY_OPERATOR_VALUE_HPP
#define EDV_BINARY_OPERATOR_VALUE_HPP

#include <edv/value.hpp>
#include <edv/implicit_value.hpp>
#include <boost/bind.hpp>

#define EDV_MAKE_BINARY_OP_VALUE(name, op) \
  template <typename Type> \
  class name : public value<Type> \
  { \
  public: \
    typedef Type type; \
    name(implicit_value<type> v1, implicit_value<type> v2) : \
      value<type>(v1.get() op v2.get()), m_value1(v1), m_value2(v2) \
    { \
      m_value1.listen_change(boost::bind(&name::on_change, this)); \
      m_value2.listen_change(boost::bind(&name::on_change, this)); \
    } \
  private: \
    void on_change() \
    { \
      set(m_value1.get() op m_value2.get()); \
    } \
    implicit_value<type> m_value1, m_value2; \
  }

namespace edv
{
  EDV_MAKE_BINARY_OP_VALUE(sum_value, +);
  EDV_MAKE_BINARY_OP_VALUE(diff_value, -);
  EDV_MAKE_BINARY_OP_VALUE(product_value, *);
  EDV_MAKE_BINARY_OP_VALUE(division_value, /);
  EDV_MAKE_BINARY_OP_VALUE(and_value, &&);
  EDV_MAKE_BINARY_OP_VALUE(or_value, ||);
  EDV_MAKE_BINARY_OP_VALUE(binary_and_value, &);
  EDV_MAKE_BINARY_OP_VALUE(binary_or_value, |);
}

#define EDV_MAKE_UNARY_OP_VALUE(name, op) \
  template <typename Type> \
  class name : public value<Type> \
  { \
  }

namespace edv {
EDV_MAKE_UNARY_OP_VALUE(not_value, !);
}


#endif  // EDV_BINARY_OPERATOR_VALUE_HPP
