#include <iostream>
#include <boost/signal.hpp>

#include <edv/value.hpp>
#include <edv/implicit_value.hpp>
#include <edv/string_values.hpp>
#include <edv/binary_operator_value.hpp>
#include <edv/sync_value.hpp>
#include <edv/relations.hpp>
#include <edv/value_view.hpp>
#include <boost/bind.hpp>

using namespace std;
using namespace edv;

template <typename Type>
class MyObserver
{
public:
  void Observe(const std::string& name, const value<Type>& v)
  {
    v.listen_change(boost::bind(&MyObserver::NotifyChange, this, &v, _1));
    mNames[&v] = name;
  }
private:
  void NotifyChange(const value<Type>* v, Type previousValue)
  {
    cout << "value " << mNames[v] << " has been changed from " << 
      previousValue << " to " << v->get() << endl;
  }
  std::map<const edv::value<Type>*, std::string> mNames;
};

int double_value(int v) { return v * 2; }

int main(int, char**)
{
  MyObserver<int> observer;
  concrete_value<int> i(4), j(5), k(6);
  observer.Observe("i", i);
  observer.Observe("j", j);
  observer.Observe("k", k);
  i = 6;
  k = 4;
  
  sum_value<int> sum(i, k);
  observer.Observe("sum", sum);
  i = 4;
  k = 6;
  
  MyObserver<std::string> strObserver;
  {
    string_convert_value<int, std::string> strJ(j);
    strObserver.Observe("strJ", strJ);
    j = 10;
  }
  ostringstream_value<std::string> str;
  strObserver.Observe("str", str);
  str << "i(" << i << ") + k(" << k << ") = " << 
    new sum_value<int>(i, k);
  i = 6;
  k = 4;
  
  sync_value<int> si(4);
  sync_value<int> sj(6);
  sync_value<int> sk(8);
  observer.Observe("si", si);
  observer.Observe("sj", sj);
  observer.Observe("sk", sk);
  make_equal_relation(si, sj);
  make_diff_relation(sj, sk, 10);
  
  si = 5;
  sync_value<int>::disconnect(sj, sk);
  si = 4;
  
  
  // examples from doc
  {
    concrete_value<int> i(4);
    value_view<int> j(i);
    value_view<int> k(j, boost::function<int (int)>(&double_value));
    i = 6;  // j now equal to 6, k to 12
    cout << i << ", " << j << ", " << k << endl;
  }
  {
    concrete_value<int> i(4), j(6);
    sum_value<int> sum(i, j);  // sum now equal to 10
    diff_value<int> diff(i, j);  // diff now equal to -2
    cout << sum << ", " << diff << endl;
    i = 6;  // sum now equal to 12, diff to 0
    cout << sum << ", " << diff << endl;
    sum_value<int> result(i, new diff_value<int>(j, 7));
    cout << result << endl;
  }
  {
    concrete_value<int> i(3);
    string_convert_value<int, std::string> str(i);  // str content is "3"
    i = 0;  // str content becomes "0"
    cout << str.get() << endl;
  }
  {
    ostringstream_value<std::string> str2;
    concrete_value<int> i(4), j(6);
    str2 << "The sum of " << i << " and " << j << " is " << 
      new sum_value<int>(i, j);
    i = 6;  // str2 content is changed accordingly
    cout << str2.get() << endl;
  }
  {
    sync_value<double> i(4);
    sync_value<double> j(6);
    sync_value<double> k(8);
    make_equal_relation(i, j);  // i and j equal to 4
    make_diff_relation(j, k, 10.0);  // k now equal to 14
    i = 2;  // j now equal to 2 and k to 12
    cout << i << ", " << j << ", " << k << endl;
    j = 3;  // i now equal to 3 and k to 13
    cout << i << ", " << j << ", " << k << endl;
    k = 4;  // i and j now equal to -6
    cout << i << ", " << j << ", " << k << endl;
  }
}

