// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_RELATIONS_HPP
#define EDV_RELATIONS_HPP

#include <edv/sync_value.hpp>
#include <boost/bind.hpp>

namespace edv
{
  namespace detail
  {
    template <typename Type>
    Type diff(Type v1, Type v2)
    {
      return v1 - v2;
    }

    template <typename Type>
    Type sum(Type v1, Type v2)
    {
      return v1 + v2;
    }
    
    template <typename Type>
    Type same(Type v)
    {
      return v;
    }
  }

  template <typename Type>
  void make_equal_relation(
    sync_value<Type>& sync1, sync_value<Type>& sync2)
  {
    connect(sync1, sync2,
       boost::function<Type (Type)>(&detail::same<Type>),
       boost::function<Type (Type)>(&detail::same<Type>));
  }

  template <typename Type>
  void make_diff_relation(
    sync_value<Type>& sync1, sync_value<Type>& sync2, Type v)
  {
    connect(sync1, sync2, 
      boost::function<Type (Type)>(boost::bind(detail::diff<Type>, _1, v)),
      boost::function<Type (Type)>(boost::bind(detail::sum<Type>, _1, v)));
  }


}  // namespace edv
#endif  // EDV_RELATIONS_HPP
