// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_VALUE_HPP
#define EDV_VALUE_HPP

#include <boost/signal.hpp>

namespace edv {
  // The value class is the base class for all event-driven values.
  template <typename Type>
  class value
  {
  public:
    typedef Type type;
    typedef boost::signal<void (type previous_value)> change_signal;
    typedef boost::signal<void ()> deletion_signal;
    
    inline type get() const
    {
      return m_real_value;
    }
    virtual ~value()
    {
      m_deletion_signal();
    }
    inline boost::signals::connection listen_change(
      const typename change_signal::slot_type slot) const
    {
      return m_change_signal.connect(slot);
    }
    inline boost::signals::connection listen_deletion(
      const typename deletion_signal::slot_type slot) const
    {
      return m_deletion_signal.connect(slot);
    }
  protected:
    explicit value(type v) : m_real_value(v) {}
    
    inline void set(type v)
    {
      if (m_real_value != v)
      {
        type previous_value = m_real_value;
        m_real_value = v;
	m_change_signal(previous_value);
      }
    }
  private:
    type m_real_value;
    mutable change_signal m_change_signal;
    mutable deletion_signal m_deletion_signal;
  };
  
  template <typename Type>
  class concrete_value : public value<Type>
  {
  public:
    typedef Type type;
    explicit concrete_value(type v) : value<type>(v) {}
    concrete_value() : value<type>(type()) {}
    using value<Type>::set;
    concrete_value<Type>& operator=(type v)
    {
      set(v);
      return *this;
    }
    concrete_value<Type>& operator=(const value<type>& v)
    {
      set(v.get());
      return *this;
    }
  };
  
  template <typename Type>
  std::ostream& operator<<(std::ostream& out, const value<Type>& v)
  {
    out << v.get();
  }
}  // namespace edv

#endif  // EDV_VALUE_HPP
