// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_VALUE_VIEW_HPP
#define EDV_VALUE_VIEW_HPP

#include <edv/value.hpp>

namespace edv {
  // The value_view class is used to build value objects having the value of
  // another value object, or having a value resulting of an operation on the
  // value of another value object.
  template <typename Type>
  class value_view : public value<Type>, public boost::signals::trackable
  {
  public:
    typedef Type type;
    value_view(type default_value, 
      boost::function<type (type)> modifier = boost::function<type (type)>()) :
      value<type>(default_value), m_modifier(modifier), m_observed_value(0)
    {
    }
    value_view(const value<type>& v, 
      boost::function<type (type)> modifier = boost::function<type (type)>()) :
      value<type>(v.get()), m_modifier(modifier), m_observed_value(0)
    {
      m_observed_value = &v;
      m_connection = m_observed_value->listen_change(
        boost::bind(&value_view<type>::on_value_change, this, _1));
    }
    void connect(const value<type>& v)
    {
      if (m_observed_value)
      {
      	disconnect();
      }
      m_observed_value = &v;
      m_connection = m_observed_value->listen_change(
        boost::bind(&value_view<type>::on_value_change, this, _1));
      set(v.get());
    }
    void disconnect(const value<type>& v)
    {
      assert(m_observed_value);
      m_connection.disconnect();
    }
  private:  
    void on_value_change(type)
    {
      if (m_modifier.empty())
      {
        set(m_observed_value->get());
      }
      else
      {
        set(m_modifier(m_observed_value->get()));
      }
    }
    boost::function<type (type)> m_modifier;
    boost::signals::connection m_connection; 
    const value<type>* m_observed_value;
  };
}  // namespace edv

#endif  // EDV_VALUE_VIEW_HPP
